from functools import singledispatch
from typing import Optional, Sequence, Tuple, cast

import cryptography
import cryptography.x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePublicKey, ECDSA, SECP256R1)
from cryptography.hazmat.primitives.hashes import (SHA256, SHA384, SHA512)
from cryptography.x509 import Certificate

from .converters import cryptography_public_key
from .errors import (AttestationError, UnimplementedError, ValidationError,
                     VerificationError)
from .types import (AndroidKeyAttestationStatement, AttestationObject,
                    AttestationStatement, AttestationType, CredentialPublicKey,
                    EC2CredentialPublicKey, FIDOU2FAttestationStatement,
                    NoneAttestationStatement, TrustedPath)


@singledispatch
def attest(att_stmt: AttestationStatement, att_obj: AttestationObject,
           auth_data: bytes,
           client_data_hash: bytes) -> Tuple[AttestationType, TrustedPath]:
  raise UnimplementedError('{} verification unimplemented'.format(
      type(att_stmt)))


@attest.register(FIDOU2FAttestationStatement)
def attest_fido_u2f(att_stmt: FIDOU2FAttestationStatement,
                    att_obj: AttestationObject, auth_data: bytes,
                    client_data_hash: bytes
                    ) -> Tuple[AttestationType, TrustedPath]:
  if len(att_stmt.x5c) != 1:
    raise ValidationError(
        'FIDO U2F verification failed: must have a single X.509 certificate')

  att_cert = att_stmt.x5c[0]
  att_cert_x509 = cryptography.x509.load_pem_x509_certificate(
      att_cert, default_backend())
  att_cert_x509_pk = att_cert_x509.public_key()
  if not isinstance(att_cert_x509_pk, EllipticCurvePublicKey):
    raise ValidationError(
        'FIDO U2F verification failed: must use an Elliptic Curve Public Key')

  if not isinstance(att_cert_x509_pk.curve, SECP256R1):
    raise ValidationError(
        'FIDO U2F verification failed: must use an Elliptic Curve Public Key')

  assert att_obj.auth_data is not None
  assert att_obj.auth_data.attested_credential_data is not None

  credential_public_key = cast(
      EC2CredentialPublicKey,
      att_obj.auth_data.attested_credential_data.credential_public_key)

  assert credential_public_key is not None
  public_key_u2f = bytes.fromhex('04') + (credential_public_key.x +
                                          credential_public_key.y)

  rp_id_hash = att_obj.auth_data.rp_id_hash
  credential_id = att_obj.auth_data.attested_credential_data.credential_id
  verification_data = (bytes.fromhex('00') + rp_id_hash +
                       (client_data_hash + credential_id + public_key_u2f))

  try:
    att_cert_x509_pk.verify(att_stmt.sig, verification_data, SHA256())
  except cryptography.exceptions.InvalidSignature:
    raise VerificationError('FIDO U2F verification failed: invalid signature')

  return AttestationType.UNCERTAIN, [att_cert_x509]


@attest.register(AndroidKeyAttestationStatement)
def attest_android_key(att_stmt: AndroidKeyAttestationStatement,
                       att_obj: AttestationObject, auth_data: bytes,
                       client_data_hash: bytes
                       ) -> Tuple[AttestationType, TrustedPath]:
  if len(att_stmt.x5c) == 0:
    raise ValidationError('Must have at least 1 X509 certificate')

  credential_certificate = cryptography.x509.load_pem_x509_certificate(
      att_stmt.x5c[0], default_backend())
  cred_cert_pk = credential_certificate.public_key()

  assert att_stmt.alg is not None
  alg_name = att_stmt.alg.name

  alg_to_hash = {
      'ES256': SHA256,
      'ES384': SHA384,
      'ES512': SHA512,
  }

  hash_algorithm = None
  if alg_name in alg_to_hash: hash_algorithm = ECDSA(alg_to_hash[alg_name]())
  elif alg_name != 'EDDSA':
    raise ValidationError('Unsupported hashing algorithm {}'.format(alg_name))

  verification_data = auth_data + client_data_hash
  try:
    if hash_algorithm is not None:
      cred_cert_pk.verify(att_stmt.sig, verification_data, hash_algorithm)
    else:
      cred_cert_pk.verify(att_stmt.sig, verification_data)
  except cryptography.exceptions.InvalidSignature:
    raise VerificationError(
        'Android Key verification failed: invalid signature')

  assert att_obj.auth_data is not None
  assert att_obj.auth_data.attested_credential_data is not None

  cpk = cryptography_public_key(
      att_obj.auth_data.attested_credential_data.credential_public_key)

  if cpk != cred_cert_pk:
    raise ValidationError(
        ('Certificate public key in attestation statement must match the '
         'provided credential public key'))

  return AttestationType.UNCERTAIN, [credential_certificate]


@attest.register(NoneAttestationStatement)
def attest_none(att_stmt: NoneAttestationStatement, att_obj: AttestationObject,
                auth_data: bytes, client_data_hash: bytes
                ) -> Tuple[AttestationType, TrustedPath]:
  return AttestationType.NONE, None



from typing import Any, NamedTuple, Optional, Sequence, Union

from .errors import UnimplementedError
from .types import (AttestationObject, AttestationStatementFormatIdentifier,
                    AttestationType, AuthenticatorAssertionResponse,
                    AuthenticatorAttestationResponse, AuthenticatorData,
                    CredentialCreationOptions, CredentialPublicKey,
                    CredentialRequestOptions, PublicKey, PublicKeyCredential,
                    PublicKeyCredentialRpEntity, PublicKeyCredentialUserEntity,
                    TrustedPath)


class CredentialData(NamedTuple):
  credential_public_key: CredentialPublicKey
  signature_count: int
  user_entity: Optional[PublicKeyCredentialUserEntity] = None
  rp_entity: Optional[PublicKeyCredentialRpEntity] = None


class CredentialsRegistrar:
  def register_creation_options(self,
                                options: CredentialCreationOptions) -> bool:
    return True

  def register_request_options(self,
                               options: CredentialRequestOptions) -> bool:
    return True

  def register_credential_creation(self,
                                   credential: PublicKeyCredential,
                                   att: AttestationObject,
                                   att_type: AttestationType,
                                   user: PublicKeyCredentialUserEntity,
                                   rp: PublicKeyCredentialRpEntity,
                                   trusted_path: Optional[TrustedPath] = None
                                   ) -> bool:
    raise UnimplementedError('Must implement register_credential_creation')

  def register_credential_request(self, credential: PublicKeyCredential,
                                  authenticator_data: AuthenticatorData,
                                  user: PublicKeyCredentialUserEntity,
                                  rp: PublicKeyCredentialRpEntity) -> bool:
    raise UnimplementedError('Must implement register_credential_request')

  def get_credential_data(self,
                          credential_id: bytes) -> Optional[CredentialData]:
    raise UnimplementedError('Must implement get_credential_data')

  def check_user_owns_credential(self, user_handle: bytes,
                                 credential_id: bytes) -> Optional[bool]:
    raise UnimplementedError('Must implement check_user_owns_credential')



import hashlib
import json
from typing import Any, Optional, Sequence, Set, cast

import cbor
import cryptography
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
from cryptography.hazmat.primitives.hashes import SHA256

from .attesters import attest
from .converters import cryptography_public_key, jsonify
from .errors import (AuthenticationError, DecodingError, IntegrityError,
                     NotFoundError, ParseError, RegistrationError,
                     SignatureCountError, ValidationError, VerificationError)
from .parsers import (parse_attestation, parse_authenticator_data,
                      parse_client_data)
from .registrars import CredentialsRegistrar
from .types import (AuthenticatorAssertionResponse,
                    AuthenticatorAttestationResponse, AuthenticatorDataFlag,
                    CredentialCreationOptions, CredentialRequestOptions,
                    ExtensionIdentifier, PublicKeyCredential,
                    PublicKeyCredentialDescriptor, PublicKeyCredentialRpEntity,
                    PublicKeyCredentialUserEntity, TokenBinding)
from .utils import extract_origin, url_base64_decode
from .verifiers import verify


class CredentialsBackend:
  def __init__(self, registrar: CredentialsRegistrar):
    self.registrar = registrar

  def handle_creation_options(self, *, options: CredentialCreationOptions):
    if not self.registrar.register_creation_options(options):
      raise RegistrationError('Failed to register creation options')

  def handle_request_options(self, *, options: CredentialRequestOptions):
    if not self.registrar.register_request_options(options):
      raise RegistrationError('Failed to register request options')

  def handle_credential_creation(
      self,
      *,
      credential: PublicKeyCredential,
      user: PublicKeyCredentialUserEntity,
      rp: PublicKeyCredentialRpEntity,
      expected_challenge: bytes,
      token_binding: Optional[TokenBinding] = None,
      require_user_verification: bool = False,
      expected_extensions: Optional[Set[ExtensionIdentifier]] = None):
    response = cast(AuthenticatorAttestationResponse, credential.response)
    collected_client_data = parse_client_data(response.client_data_JSON)
    if collected_client_data is None:
      raise ParseError('Could not parse the client data dictionary')

    if collected_client_data.type != 'webauthn.create':
      raise ValidationError('Invalid client data type {}'.format(
          collected_client_data.type))

    try:
      challenge = url_base64_decode(collected_client_data.challenge)
      if challenge != expected_challenge:
        raise IntegrityError(
            'Given and expected challenge byte strings don\'t match')
    except ValueError:
      raise DecodingError('Failed to decode the base64 encoded challenge')

    rp_origin = rp_origin = extract_origin(collected_client_data.origin)
    if rp_origin != rp.id:
      raise IntegrityError(
          'Given ({0}) and expected ({1}) RP origin don\'t match'.format(
              rp_origin, rp.id))

    if token_binding is not None:
      if collected_client_data.token_binding is None:
        raise IntegrityError('Expecting Token Binding in client data')

      if token_binding.status != collected_client_data.token_binding.status:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'statuses dont\'t match').format(
                                  collected_client_data.token_binding.status,
                                  token_binding.status))

      if token_binding.id != collected_client_data.token_binding.id:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'ids dont\'t match').format(
                                  collected_client_data.token_binding.id,
                                  token_binding.id))
    elif collected_client_data.token_binding is not None:
      raise IntegrityError('Unexpected Token Binding in client data')

    client_data_JSON_hash = hashlib.sha256(response.client_data_JSON).digest()
    rp_id_hash = hashlib.sha256(rp.id.encode('utf-8')).digest()

    attestation, raw_att_obj = parse_attestation(response.attestation_object)
    if attestation.auth_data.rp_id_hash != rp_id_hash:
      raise IntegrityError('RP ID hash does not match')

    if not (AuthenticatorDataFlag.UP.value & attestation.auth_data.flags):
      raise ValidationError('User present flag must be set')

    if require_user_verification:
      if not (AuthenticatorDataFlag.UV.value & attestation.auth_data.flags):
        raise ValidationError('User verification flag must be set')

    if expected_extensions is not None:
      for e in expected_extensions:
        if not hasattr(attestation.auth_data.extensions, e.key) or getattr(
            attestation.auth_data.extensions, e.key) is None:
          raise ValidationError('Missing extension {}'.format(e.value))

    assert attestation.auth_data is not None
    assert attestation.auth_data.attested_credential_data is not None

    att_type, trusted_path = attest(
        attestation.att_stmt,
        attestation.auth_data.attested_credential_data.credential_public_key,
        raw_att_obj['authData'], client_data_JSON_hash)

    crypto_pk = cryptography_public_key(
        attestation.auth_data.attested_credential_data.credential_public_key)

    if not self.registrar.register_credential_creation(
        credential=credential,
        att=attestation,
        att_type=att_type,
        user=user,
        rp=rp,
        trusted_path=trusted_path,
    ):
      raise RegistrationError('Failed to register credential creation')

  def handle_credential_request(
      self,
      *,
      credential: PublicKeyCredential,
      expected_challenge: bytes,
      rp: Optional[PublicKeyCredentialRpEntity] = None,
      user: Optional[PublicKeyCredentialUserEntity] = None,
      allow_credentials: Optional[
          Sequence[PublicKeyCredentialDescriptor]] = None,
      token_binding: Optional[TokenBinding] = None,
      require_user_verification: bool = False,
      expected_extensions: Optional[Set[ExtensionIdentifier]] = None,
      ignore_clone_error: bool = False):
    response = cast(AuthenticatorAssertionResponse, credential.response)
    if allow_credentials is not None:
      allowed = False
      for cred in allow_credentials:
        if cred.id == credential.raw_id:
          allowed = True
          break

      if not allowed:
        raise ValidationError('Provided credential is not allowed')

    credential_data = self.registrar.get_credential_data(credential.raw_id)

    if credential_data is None:
      raise NotFoundError('Could not get credential data')

    registered_user = credential_data.user_entity
    if registered_user is None:
      if user is None:
        raise NotFoundError('Could not find a user with the credential')
    elif user is not None:
      if registered_user.id != user.id:
        raise ValidationError(
            'Registered user and provided user ids do not match')
    else:
      user = registered_user

    registered_rp = credential_data.rp_entity
    if registered_rp is None:
      if rp is None:
        raise NotFoundError(
            ('Could not find a registered rp with the credential '
             'and one was not provided for verification'))
    elif rp is not None:
      if registered_rp.id != rp.id:
        raise ValidationError('Registered rp and provided rp ids do not match')
    else:
      rp = registered_rp

    if user is not None:
      if response.user_handle is not None:
        if response.user_handle != user.id:
          raise ValidationError('User handle doesn\'t match user id')
    else:
      if response.user_handle is None:
        raise ValidationError(('User cannot be verified, must provide one or '
                               'there must be a user handle'))

      valid_credential = self.registrar.check_user_owns_credential(
          response.user_handle, credential.raw_id)

      if not valid_credential:
        raise ValidationError('User does not own credential')

    collected_client_data = parse_client_data(response.client_data_JSON)
    if collected_client_data is None:
      raise ParseError('Could not parse the client data dictionary')

    if collected_client_data.type != 'webauthn.get':
      raise ValidationError('Invalid client data type {}'.format(
          collected_client_data.type))

    try:
      challenge = url_base64_decode(collected_client_data.challenge)
      if challenge != expected_challenge:
        raise IntegrityError(
            'Given and expected challenge byte strings don\'t match')
    except ValueError:
      raise DecodingError('Failed to decode the base64 encoded challenge')

    rp_origin = extract_origin(collected_client_data.origin)
    if rp_origin != rp.id:
      raise IntegrityError(
          'Given ({0}) and expected ({1}) RP origin don\'t match'.format(
              rp_origin, rp.id))

    if token_binding is not None:
      if collected_client_data.token_binding is None:
        raise IntegrityError('Expecting Token Binding in client data')

      if token_binding.status != collected_client_data.token_binding.status:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'statuses dont\'t match').format(
                                  collected_client_data.token_binding.status,
                                  token_binding.status))

      if token_binding.id != collected_client_data.token_binding.id:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'ids dont\'t match').format(
                                  collected_client_data.token_binding.id,
                                  token_binding.id))
    elif collected_client_data.token_binding is not None:
      raise IntegrityError('Unexpected Token Binding in client data')

    client_data_JSON_hash = hashlib.sha256(response.client_data_JSON).digest()
    rp_id_hash = hashlib.sha256(rp.id.encode('utf-8')).digest()

    auth_data = parse_authenticator_data(response.authenticator_data)

    if auth_data.rp_id_hash != rp_id_hash:
      raise IntegrityError('RP ID hash does not match')

    if not (AuthenticatorDataFlag.UP.value & auth_data.flags):
      raise ValidationError('User present flag must be set')

    if require_user_verification:
      if not (AuthenticatorDataFlag.UV.value & auth_data.flags):
        raise ValidationError('User verification flag must be set')

    if expected_extensions is not None:
      for e in expected_extensions:
        if not hasattr(auth_data.extensions,
                       e.key) or getattr(auth_data.extensions, e.key) is None:
          raise ValidationError('Missing extension {}'.format(e.value))

    verification_data = response.authenticator_data + client_data_JSON_hash
    verify(credential_data.credential_public_key, response.signature,
           verification_data)

    registered_sign_count = credential_data.signature_count

    if auth_data.sign_count != 0 or registered_sign_count != 0:
      if auth_data.sign_count <= registered_sign_count:
        if not ignore_clone_error:
          raise SignatureCountError('Detected a possible authenticator clone')

    if not self.registrar.register_credential_request(
        credential=credential,
        authenticator_data=auth_data,
        user=user,
        rp=rp,
    ):
      raise RegistrationError('Failed to register credential request')



import setuptools
from distutils.core import setup
from os.path import dirname, join


setup(
  name='webauthn-rp',
  version='0.0.2',
  description='Web Authentication Relying Party Library',
  author='enceladus-rex',
  packages=['webauthn_rp'],
  url='https://github.com/enceladus-rex/webauthn-rp',
  long_description=open(join(dirname(__file__), 'README.md')).read(),
  long_description_content_type='text/markdown',
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  install_requires=[
    'cbor==1.0.0',
    'cryptography==2.6.1',
  ],
)
